/*
 * z2-Environment
 * 
 * Copyright(c) ZFabrik Software GmbH & Co. KG
 * 
 * contact@zfabrik.de
 * 
 * http://www.z2-environment.eu
 */
package com.zfabrik.impl.work;

import java.util.logging.Level;
import java.util.logging.Logger;

import com.zfabrik.util.function.ThrowingSupplier;

/*
 * A runnable that organizes the execution of another runnable
 * by another thread
 */
public class ThreadExecutionGate implements Runnable {
	private static final Logger LOG = Logger.getLogger(ThreadExecutionGate.class.getName());
	private static final short BLOCKED = 0;
	private static final short OPEN = 1;
	private static final short DONE = 2;
	
	private short status = BLOCKED;
	
	/**
	 * When run, let the (possibly waiting) other thread in {@link #work(ThrowingSupplier)} pass 
	 */
	@Override
	public void run() {
		synchronized (this) {
			// open the gate
			this.status = OPEN;
			this.notifyAll();
			// wait until work completed
			while (this.status!=DONE) {
				try {
					this.wait();
				} catch (InterruptedException e) {
					LOG.log(Level.SEVERE, "RunAs gateway wait for done interrupted",e);
				}
			}
		}
		
	}
	
	/**
	 * Do the actual thing but have it wait to {@link #run()}.
	 */
	public <T, E extends Exception> T work(ThrowingSupplier<T,E> supplier) throws E {
		synchronized (this) {
			while (this.status==BLOCKED) {
				// if still blocked, wait until open
				try {
					this.wait();
				} catch (InterruptedException e) {
					LOG.log(Level.SEVERE, "RunAs gateway wait for pass interrupted",e);
				}
			}
		}
		try {
			return supplier.get();
		} finally {
			synchronized (this) {
				// indicate completion
				this.status=DONE;
				this.notify();
			}
		}
	}

}
