/*
 * z2-Environment
 *
 * Copyright(c) ZFabrik Software KG
 *
 * contact@zfabrik.de
 *
 * http://www.z2-environment.eu
 */
package com.zfabrik.impl.db.data;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLTimeoutException;
import java.sql.Statement;
import java.time.Duration;
import java.time.Instant;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Objects;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Collectors;

/**
 * Supports performance tracing of JDBC provided objects from our pool.
 * <p>
 * Just set the JUL logger "performance.com.zfabrik.impl.db.data" to at least logging level FINE and it
 * will output statement operation performance data.
 * <p>
 * Use log level:
 * <OL>
 * <li>FINE     -> only expensive queries are logged</li>
 * <li>FINER    -> all statements are logged</li>
 * <li>FINEST   -> all statements plus statement details are logged</li>
 * </OL>
 */
public class JDBCTracer {

    private final static Logger      LOGGER                         = Logger.getLogger("performance." + JDBCTracer.class.getName());
    private final static long        EXPENSIVE_QUERY_THRESHOLD_MSEC = 250;
    private static final Set<String> IGNORE_METHODS                 = new HashSet<>(Arrays.asList("hashCode", "equals"));

    public static Connection treatThis(Connection conn) {
        if (LOGGER.isLoggable(Level.FINE)) {
            return (Connection) Proxy.newProxyInstance(
                    Connection.class.getClassLoader(),
                    new Class[]{Connection.class},
                    new ConnectionInvocationHandler(conn)
            );
        }
        return conn;
    }

    // anything traceable in here...
    private static class Traceable {
        private Instant  t0;
        private Long     dtMillis;
        private String   methodName;
        private Object[] args;
        private boolean  success;
        private Level    minLevel = Level.FINER;

        void begin(Method method, Object[] args) {
            this.t0 = Instant.now();
            this.dtMillis = null;
            this.methodName = method.getName();
            this.args = args;
        }

        void finished() {
            if (LOGGER.isLoggable(minLevel)) {
                LOGGER.log(minLevel, toString() + " Duration = " + getDurationMillis() + " msec");
            }
        }

        void setRequiredLevel(Level level) {
            minLevel = level;
        }

        void setSuccess() {
            this.success = true;
        }

        long getDurationMillis() {
            if (dtMillis == null) {
                dtMillis = Duration.between(t0, Instant.now()).toMillis();
            }
            return dtMillis;
        }

        @Override
        public String toString() {
            return (success ? "" : "FAILED ") +
                   methodName +
                   (args == null ? "(): " : Arrays.stream(args).map(Objects::toString).collect(Collectors.joining(", ", "(", ") ")));
        }
    }

    private static class StatementTraceable extends Traceable {
        private final Statement statement;

        private StatementTraceable(Statement statement) {this.statement = statement;}

        Statement getStatement() {
            return statement;
        }

        @Override
        void finished() {
            super.finished();

            if (getDurationMillis() > 2 * EXPENSIVE_QUERY_THRESHOLD_MSEC) {
                // log expensive queries in any case
                LOGGER.log(
                        Level.WARNING,
                        "Very Expensive Query detected",
                        new SQLTimeoutException(toString() + "\nDuration = " + getDurationMillis() + " msec")
                );

            } else if (getDurationMillis() > EXPENSIVE_QUERY_THRESHOLD_MSEC) {
                LOGGER.log(
                        Level.INFO,
                        "Expensive Query " + toString() + "\nDuration = " + getDurationMillis() + " msec"
                );
            }
        }
    }

    // invocation handler for connections
    //
    private static class ConnectionInvocationHandler extends Traceable implements InvocationHandler {
        private Connection connection;

        public ConnectionInvocationHandler(Connection connection) {
            this.connection = connection;
        }

        /**
         * @see java.lang.reflect.InvocationHandler#invoke(Object, Method, Object[])
         */
        @Override
        public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
            if (IGNORE_METHODS.contains(method.getName())) {
                return method.invoke(connection, args);
            }

            begin(method, args);
            try {
                Object obj = method.invoke(connection, args);
                if ("createStatement".equals(method.getName())) {
                    obj = Proxy.newProxyInstance(Statement.class.getClassLoader(),
                                                 new Class[]{Statement.class},
                                                 new StatementInvocationHandler((Statement) obj)
                    );
                } else if ("prepareStatement".equals(method.getName())) {
                    obj = Proxy.newProxyInstance(PreparedStatement.class.getClassLoader(),
                                                 new Class[]{PreparedStatement.class},
                                                 new StatementInvocationHandler((PreparedStatement) obj)
                    );
                }
                setSuccess();
                return obj;
            } finally {
                finished();
            }
        }
    }

    //
    // begin class StatementWrapper
    //
    private static class StatementInvocationHandler extends StatementTraceable implements InvocationHandler {

        public StatementInvocationHandler(Statement statement) {
            super(statement);
        }

        /**
         * @see java.lang.reflect.InvocationHandler#invoke(Object, Method, Object[])
         */
        @Override
        public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {

            if (IGNORE_METHODS.contains(method.getName())) {
                // shortcut for methods which do not concern us
                return method.invoke(getStatement(), args);
            }

            String methodName = method.getName();

            // log calls != execute* only if log level is FINEST
            setRequiredLevel(
                    methodName.startsWith("execute")
                    ? Level.FINER
                    : Level.FINEST
            );

            begin(method, args);
            try {
                Object obj = method.invoke(getStatement(), args);
                setSuccess();
                return obj;
            } finally {
                finished();
            }
        }
    }
}
