/*
 * z2-Environment
 * 
 * Copyright(c) ZFabrik Software GmbH & Co. KG
 * 
 * contact@zfabrik.de
 * 
 * http://www.z2-environment.eu
 */
package com.zfabrik.dev.z2jupiter.internal.transfer;

import java.io.Serializable;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.file.Paths;
import java.util.Collections;
import java.util.List;
import java.util.Optional;

import org.junit.platform.engine.DiscoverySelector;
import org.junit.platform.engine.discovery.ClassSelector;
import org.junit.platform.engine.discovery.ClasspathResourceSelector;
import org.junit.platform.engine.discovery.ClasspathRootSelector;
import org.junit.platform.engine.discovery.DirectorySelector;
import org.junit.platform.engine.discovery.DiscoverySelectors;
import org.junit.platform.engine.discovery.FilePosition;
import org.junit.platform.engine.discovery.FileSelector;
import org.junit.platform.engine.discovery.MethodSelector;
import org.junit.platform.engine.discovery.ModuleSelector;
import org.junit.platform.engine.discovery.NestedClassSelector;
import org.junit.platform.engine.discovery.NestedMethodSelector;
import org.junit.platform.engine.discovery.PackageSelector;
import org.junit.platform.engine.discovery.UniqueIdSelector;
import org.junit.platform.engine.discovery.UriSelector;

/**
 * Serializable selector
 */
public class Z2JupiterSelectorDto implements Serializable {
	private static final long serialVersionUID = 2231839170622497448L;


	public static enum SelectorType {
		ClasspathResourceSelector, ClasspathRootSelector, ClassSelector, DirectorySelector, FileSelector,
		MethodSelector, ModuleSelector, NestedClassSelector, NestedMethodSelector, PackageSelector, UniqueIdSelector, UriSelector
	}

	private SelectorType type;
	private String classpathResourceName;
	private Integer line;
	private Integer column;
	private String  uri;
	private String className;
	private String nestedClassName;
	private String file;
	private String methodName;
	private String methodParameterTypes;
	private String moduleName;
	private List<String> enclosingClassNames;
	private String packageName;
	private String uniqueId;
	
	public Z2JupiterSelectorDto() {
	}

	public Z2JupiterSelectorDto(DiscoverySelector selector) {
		if (selector instanceof ClasspathResourceSelector) {
			this.type = SelectorType.ClasspathResourceSelector;
			ClasspathResourceSelector sel = (ClasspathResourceSelector) selector;
			this.classpathResourceName = sel.getClasspathResourceName();
			this.position(sel.getPosition());
		} else
		if (selector instanceof ClasspathRootSelector) {
			this.type = SelectorType.ClasspathRootSelector;
			ClasspathRootSelector sel = (ClasspathRootSelector) selector;
			this.uri = sel.getClasspathRoot().toString();
		} else
		if (selector instanceof ClassSelector) {
			this.type = SelectorType.ClassSelector;
			ClassSelector sel = (ClassSelector) selector;
			this.className = sel.getClassName();
		} else
		if (selector instanceof DirectorySelector) {
			this.type = SelectorType.DirectorySelector;
			DirectorySelector sel = (DirectorySelector) selector;
			file = sel.getDirectory().toString();
		} else
		if (selector instanceof FileSelector) {
			this.type = SelectorType.FileSelector;
			FileSelector sel = (FileSelector) selector;
			this.file = sel.getRawPath();
			this.position(sel.getPosition());
		} else
		if (selector instanceof MethodSelector) {
			this.type = SelectorType.MethodSelector;
			MethodSelector sel = (MethodSelector) selector;
			this.className=sel.getClassName();
			this.methodName=sel.getMethodName();
			this.methodParameterTypes=sel.getParameterTypeNames();
		} else
		if (selector instanceof ModuleSelector) {
			this.type = SelectorType.ModuleSelector;
			ModuleSelector sel = (ModuleSelector) selector;
			this.moduleName = sel.getModuleName();
		} else
		if (selector instanceof NestedClassSelector) {
			this.type = SelectorType.NestedClassSelector;
			NestedClassSelector sel = (NestedClassSelector) selector;
			this.enclosingClassNames = sel.getEnclosingClassNames();
			this.nestedClassName = sel.getNestedClassName();
		} else
		if (selector instanceof NestedMethodSelector) {
			this.type = SelectorType.NestedMethodSelector;
			NestedMethodSelector sel = (NestedMethodSelector) selector;
			this.enclosingClassNames=sel.getEnclosingClassNames();
			this.methodName=sel.getMethodName();
			this.methodParameterTypes=sel.getParameterTypeNames();
			this.nestedClassName=sel.getNestedClassName();
		} else
		if (selector instanceof PackageSelector) {
			this.type = SelectorType.PackageSelector;
			PackageSelector sel = (PackageSelector) selector;
			this.packageName = sel.getPackageName();
		} else
		if (selector instanceof UniqueIdSelector) {
			this.type = SelectorType.UniqueIdSelector;
			UniqueIdSelector sel = (UniqueIdSelector) selector;
			this.uniqueId = sel.getUniqueId().toString();
		} else
		if (selector instanceof UriSelector) {
			this.type = SelectorType.UriSelector;
			UriSelector sel = (UriSelector) selector;
			this.uri = sel.getUri().toString();
		} else {
			throw new IllegalArgumentException("Unsupported selector type "+selector.getClass().getName());
		}
	}

	public DiscoverySelector toSelector() {
		switch (this.type) {
		case ClasspathResourceSelector:
			return DiscoverySelectors.selectClasspathResource(this.classpathResourceName, position());
		case ClasspathRootSelector:
			try {
				return DiscoverySelectors.selectClasspathRoots(Collections.singleton(Paths.get(new URI(this.uri)))).stream().findAny().orElse(null);
			} catch (URISyntaxException e) {
				throw new IllegalArgumentException(e);
			}
		case ClassSelector:
			return DiscoverySelectors.selectClass(this.className);
		case DirectorySelector:
			return DiscoverySelectors.selectDirectory(this.file);
		case FileSelector:
			return DiscoverySelectors.selectFile(this.file,position());
		case MethodSelector:
			return DiscoverySelectors.selectMethod(this.className,  this.methodName,  this.methodParameterTypes);
		case ModuleSelector:
			return DiscoverySelectors.selectModule(this.moduleName);
		case NestedClassSelector:
			return DiscoverySelectors.selectNestedClass(this.enclosingClassNames, this.nestedClassName);
		case NestedMethodSelector:
			return DiscoverySelectors.selectNestedMethod(this.enclosingClassNames, this.nestedClassName, this.methodName, this.methodParameterTypes);
		case PackageSelector:
			return DiscoverySelectors.selectPackage(this.packageName);
		case UniqueIdSelector:
			return DiscoverySelectors.selectUniqueId(this.uniqueId);
		case UriSelector:
			return DiscoverySelectors.selectUri(this.uri);
		default:
			throw new IllegalStateException();
		}
	}
		
	private void position(Optional<FilePosition> fpos) {
		fpos.ifPresent(fp->{
			this.line = fp.getLine();
			fp.getColumn().ifPresent(co->this.column=co);
		});
	}
	
	private FilePosition position() {
		if (line!=null) {
			if (column!=null) {
				return FilePosition.from(line, column);
			} else {
				return FilePosition.from(line);
			}
		} else {
			return null;
		}
	}


	public SelectorType getType() {
		return type;
	}


	public void setType(SelectorType type) {
		this.type = type;
	}


	public String getClasspathResourceName() {
		return classpathResourceName;
	}


	public void setClasspathResourceName(String classpathResourceName) {
		this.classpathResourceName = classpathResourceName;
	}


	public Integer getLine() {
		return line;
	}


	public void setLine(Integer line) {
		this.line = line;
	}


	public Integer getColumn() {
		return column;
	}


	public void setColumn(Integer column) {
		this.column = column;
	}


	public String getUri() {
		return uri;
	}


	public void setUri(String uri) {
		this.uri = uri;
	}


	public String getClassName() {
		return className;
	}


	public void setClassName(String className) {
		this.className = className;
	}


	public String getNestedClassName() {
		return nestedClassName;
	}


	public void setNestedClassName(String nestedClassName) {
		this.nestedClassName = nestedClassName;
	}


	public String getFile() {
		return file;
	}


	public void setFile(String file) {
		this.file = file;
	}


	public String getMethodName() {
		return methodName;
	}


	public void setMethodName(String methodName) {
		this.methodName = methodName;
	}


	public String getMethodParameterTypes() {
		return methodParameterTypes;
	}


	public void setMethodParameterTypes(String methodParameterTypes) {
		this.methodParameterTypes = methodParameterTypes;
	}


	public String getModuleName() {
		return moduleName;
	}


	public void setModuleName(String moduleName) {
		this.moduleName = moduleName;
	}


	public List<String> getEnclosingClassNames() {
		return enclosingClassNames;
	}


	public void setEnclosingClassNames(List<String> enclosingClassNames) {
		this.enclosingClassNames = enclosingClassNames;
	}


	public String getPackageName() {
		return packageName;
	}


	public void setPackageName(String packageName) {
		this.packageName = packageName;
	}


	public String getUniqueId() {
		return uniqueId;
	}


	public void setUniqueId(String uniqueId) {
		this.uniqueId = uniqueId;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((className == null) ? 0 : className.hashCode());
		result = prime * result + ((classpathResourceName == null) ? 0 : classpathResourceName.hashCode());
		result = prime * result + ((column == null) ? 0 : column.hashCode());
		result = prime * result + ((enclosingClassNames == null) ? 0 : enclosingClassNames.hashCode());
		result = prime * result + ((file == null) ? 0 : file.hashCode());
		result = prime * result + ((line == null) ? 0 : line.hashCode());
		result = prime * result + ((methodName == null) ? 0 : methodName.hashCode());
		result = prime * result + ((methodParameterTypes == null) ? 0 : methodParameterTypes.hashCode());
		result = prime * result + ((moduleName == null) ? 0 : moduleName.hashCode());
		result = prime * result + ((nestedClassName == null) ? 0 : nestedClassName.hashCode());
		result = prime * result + ((packageName == null) ? 0 : packageName.hashCode());
		result = prime * result + ((type == null) ? 0 : type.hashCode());
		result = prime * result + ((uniqueId == null) ? 0 : uniqueId.hashCode());
		result = prime * result + ((uri == null) ? 0 : uri.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		Z2JupiterSelectorDto other = (Z2JupiterSelectorDto) obj;
		if (className == null) {
			if (other.className != null)
				return false;
		} else if (!className.equals(other.className))
			return false;
		if (classpathResourceName == null) {
			if (other.classpathResourceName != null)
				return false;
		} else if (!classpathResourceName.equals(other.classpathResourceName))
			return false;
		if (column == null) {
			if (other.column != null)
				return false;
		} else if (!column.equals(other.column))
			return false;
		if (enclosingClassNames == null) {
			if (other.enclosingClassNames != null)
				return false;
		} else if (!enclosingClassNames.equals(other.enclosingClassNames))
			return false;
		if (file == null) {
			if (other.file != null)
				return false;
		} else if (!file.equals(other.file))
			return false;
		if (line == null) {
			if (other.line != null)
				return false;
		} else if (!line.equals(other.line))
			return false;
		if (methodName == null) {
			if (other.methodName != null)
				return false;
		} else if (!methodName.equals(other.methodName))
			return false;
		if (methodParameterTypes == null) {
			if (other.methodParameterTypes != null)
				return false;
		} else if (!methodParameterTypes.equals(other.methodParameterTypes))
			return false;
		if (moduleName == null) {
			if (other.moduleName != null)
				return false;
		} else if (!moduleName.equals(other.moduleName))
			return false;
		if (nestedClassName == null) {
			if (other.nestedClassName != null)
				return false;
		} else if (!nestedClassName.equals(other.nestedClassName))
			return false;
		if (packageName == null) {
			if (other.packageName != null)
				return false;
		} else if (!packageName.equals(other.packageName))
			return false;
		if (type != other.type)
			return false;
		if (uniqueId == null) {
			if (other.uniqueId != null)
				return false;
		} else if (!uniqueId.equals(other.uniqueId))
			return false;
		if (uri == null) {
			if (other.uri != null)
				return false;
		} else if (!uri.equals(other.uri))
			return false;
		return true;
	}

	@Override
	public String toString() {
		return "Z2JupiterSelectorDto [type=" + type + ", classpathResourceName=" + classpathResourceName + ", line=" + line + ", column=" + column
				+ ", uri=" + uri + ", className=" + className + ", nestedClassName=" + nestedClassName + ", file=" + file + ", methodName="
				+ methodName + ", methodParameterTypes=" + methodParameterTypes + ", moduleName=" + moduleName + ", enclosingClassNames="
				+ enclosingClassNames + ", packageName=" + packageName + ", uniqueId=" + uniqueId + "]";
	}

	
}
