/*
 * z2-Environment
 * 
 * Copyright(c) ZFabrik Software GmbH & Co. KG
 * 
 * contact@zfabrik.de
 * 
 * http://www.z2-environment.eu
 */
package com.zfabrik.test.tx;

import javax.transaction.Status;

import org.junit.Assert;
import org.junit.Test;

import com.zfabrik.impl.tx.UserTransactionImpl;
import com.zfabrik.tx.TransactionUtil;
import com.zfabrik.tx.TransactionUtil.TransactionUtilException;
import com.zfabrik.work.WorkUnit;

/**
 * Unit tests for JTA support
 */
public class TransactionUtilTests {
	private UserTransactionImpl ut = new UserTransactionImpl();

	/**
	 * rollback only 
	 */
	@Test
	public void nestedTransactionRollbackOnly() throws Exception {
		TransactionUtil.run(ut,()->{
			TransactionUtil.run(ut,()->{
				ut.setRollbackOnly();
				Assert.assertEquals(Status.STATUS_MARKED_ROLLBACK, ut.getStatus());
			});
			Assert.assertEquals(Status.STATUS_MARKED_ROLLBACK, ut.getStatus());
		});
		Assert.assertEquals(Status.STATUS_NO_TRANSACTION, ut.getStatus());
	}
	
	/**
	 * rollback removes tx and so {@link TransactionUtil} freaks out.
	 */
	@Test
	public void transactionRollback() throws Exception {
		try {
			TransactionUtil.run(ut,()->{
				ut.rollback();
				Assert.assertEquals(Status.STATUS_NO_TRANSACTION, ut.getStatus());
			});
		} catch (TransactionUtilException tue) {
			Assert.assertTrue(tue.getCause() instanceof IllegalStateException);
			Assert.assertTrue(tue.getCause().getMessage().contains("No transaction associated with thread"));
		}
	}
	
	/**
	 * rollback removes tx and so {@link TransactionUtil} freaks out.
	 */
	@Test
	public void transactionCommit() throws Exception {
		try {
			TransactionUtil.run(ut,()->{
				ut.commit();
				Assert.assertEquals(Status.STATUS_NO_TRANSACTION, ut.getStatus());
			});
		} catch (TransactionUtilException tue) {
			Assert.assertTrue(tue.getCause() instanceof IllegalStateException);
			Assert.assertTrue(tue.getCause().getMessage().contains("No transaction associated with thread"));
		}
	}
	
	/**
	 * never passes outside of transaction, fails inside transaction 
	 */
	@Test
	public void transactionNever() throws Exception {
		// ok
		TransactionUtil.never(ut);
		try {
			TransactionUtil.run(ut,()->{
				TransactionUtil.never(ut);
			});
		} catch (IllegalStateException ise) {
			Assert.assertTrue(ise.getMessage().contains("No ongoing transaction allowed"));
		}
	}

	/**
	 * mandatory fails outside of transaction, passes inside transaction 
	 */
	@Test
	public void transactionMandatory() throws Exception {
		// ok
		try {
			TransactionUtil.mandatory(ut);
		} catch (IllegalStateException ise) {
			Assert.assertTrue(ise.getMessage().contains("An ongoing transaction is required"));
		}
		TransactionUtil.run(ut,()->{
			TransactionUtil.mandatory(ut);
		});
	}
	
	/**
	 * Testing that rollbackonly is reset after rollback in transaction in workunit
	 */
	@Test
	public void rollbackOnlyResetAfterTransaction() throws Exception {
		TransactionUtil.run(ut,()->{
			ut.setRollbackOnly();
		});
		TransactionUtil.run(ut,()->{
			Assert.assertFalse(ut.isRollbackOnly());
			Assert.assertFalse(WorkUnit.getRollbackOnlyCurrent());
		});
	}

	/**
	 * Testing that rollbackonly is reset after rollback in transaction in workunit
	 */
	@Test
	public void rollbackOnlyResetAfterTransactionException() throws Exception {
		WorkUnit.work(()->{
			try {
				TransactionUtil.get(ut,()->{
					throw new IllegalStateException();
				});
				Assert.fail();
			} catch (IllegalStateException re) {
				// ok
			}
			TransactionUtil.run(ut,()->{
				Assert.assertFalse(ut.isRollbackOnly());
				Assert.assertFalse(WorkUnit.getRollbackOnlyCurrent());
			});
			return null;
		});
	}


}
